@@ -1,6 +1,6 @@
 #!/bin/sh
 #*******************************************************************************
-# Copyright (c) 2000, 2009 IBM Corporation and others.
+# Copyright (c) 2000, 2010 IBM Corporation and others.
 # All rights reserved. This program and the accompanying materials
 # are made available under the terms of the Eclipse Public License v1.0
 # which accompanies this distribution, and is available at
@@ -29,12 +29,22 @@ if [ "${OS}" = "" ]; then
 	OS=`uname -s`
 fi
 case $OS in
+	"AIX")
+		SWT_OS=aix
+		MAKEFILE=make_aix.mak
+		;;
+	"HP-UX")
+		SWT_OS=hpux
+		MAKEFILE=make_hpux.mak
+		;;
 	"SunOS")
 		SWT_OS=solaris
 		PROC=`uname -i`
 		MAKEFILE=make_solaris.mak
-		if uname -p > /dev/null 2>&1; then
-			MODEL=`uname -p`
+		if [ "${MODEL}" = "" ]; then
+			if uname -p > /dev/null 2>&1; then
+				MODEL=`uname -p`
+			fi
 		fi
 		if [ ${MODEL} = 'i386' ]; then
 			MAKEFILE=make_solaris_x86.mak
@@ -55,6 +65,9 @@ esac
 if [ "${MODEL}" = "" ]; then
 	if uname -i > /dev/null 2>&1; then
 		MODEL=`uname -i`
+		if [ ${MODEL} = 'unknown' ]; then
+		  MODEL=`uname -m`
+		fi
 	else
 		MODEL=`uname -m`
 	fi
@@ -64,6 +77,10 @@ case $MODEL in
 		SWT_ARCH=x86_64
 		AWT_ARCH=amd64
 		;;
+	"sparc64")
+		SWT_ARCH=$MODEL
+		AWT_ARCH=sparcv9
+		;;
 	i?86)
 		SWT_ARCH=x86
 		AWT_ARCH=i386
@@ -73,18 +90,48 @@ case $MODEL in
 		AWT_ARCH=$MODEL
 		;;
 esac
-
+echo "Model is ${MODEL}"
 # For 64-bit CPUs, we have a switch
-if [ ${MODEL} = 'x86_64' -o ${MODEL} = 'ppc64' -o ${MODEL} = 'ia64' -o ${MODEL} = 's390x' ]; then
+if [ ${MODEL} = 'x86_64' -o ${MODEL} = 'ppc64' -o ${MODEL} = 'ia64' -o ${MODEL} = 'sparc64'  -o ${MODEL} = 's390x' ]; then
 	SWT_PTR_CFLAGS=-DJNI64
-	export SWT_PTR_CFLAGS
 	if [ -d /lib64 ]; then
 		XLIB64=-L/usr/X11R6/lib64
 		export XLIB64
 	fi
+	if [ ${MODEL} = 'ppc64' ]; then
+		if [ ${OS} = 'AIX' ]; then
+			SWT_PTR_CFLAGS="${SWT_PTR_CFLAGS} -maix64"
+			SWT_LFLAGS=-maix64
+			export SWT_LFLAGS
+		else
+			SWT_PTR_CFLAGS="${SWT_PTR_CFLAGS} -m64"	
+			XLIB64="${XLIB64} -L/usr/lib64"
+			SWT_LFLAGS=-m64
+			export SWT_LFLAGS
+		fi
+	fi
+	if [ ${MODEL} = 'ia64' ]; then
+		if [ ${OS} = 'HP-UX' ]; then
+			SWT_PTR_CFLAGS="${SWT_PTR_CFLAGS} -mlp64"
+			SWT_LFLAGS=-mlp64
+			export SWT_LFLAGS
+		fi
+	fi
+	if [ ${MODEL} = 'sparc64' ]; then
+			SWT_PTR_CFLAGS="${SWT_PTR_CFLAGS} -xarch=v9"
+			SWT_LFLAGS="-xarch=v9"
+			SWT_CDE_64SUFFIX="/64"
+			export SWT_LFLAGS SWT_CDE_64SUFFIX
+	fi
+	export SWT_PTR_CFLAGS
+fi
+if [ ${MODEL} = 's390' ]; then
+	SWT_PTR_CFLAGS="-m31"	
+	SWT_LFLAGS=-m31
+	export SWT_LFLAGS SWT_PTR_CFLAGS
 fi
 
-if [ x`pkg-config --exists gnome-vfs-module-2.0 libgnome-2.0 libgnomeui-2.0 && echo YES` = "xYES" ]; then
+if [ x`pkg-config --exists gnome-vfs-module-2.0 libgnome-2.0 libgnomeui-2.0 && echo YES` = "xYES"  -a 	 ${MODEL} != "sparc64" 	]; then
 	echo "libgnomeui-2.0 found, compiling SWT program support using GNOME"
 	MAKE_GNOME=make_gnome
 else
@@ -100,7 +147,7 @@ else
 	echo "    *** Advanced graphics support using cairo will not be compiled."
 fi
 
-if [ -z "${MOZILLA_INCLUDES}" -a -z "${MOZILLA_LIBS}" ]; then
+if [ -z "${MOZILLA_INCLUDES}" -a -z "${MOZILLA_LIBS}" -a ${MODEL} != 'sparc64' ]; then
 	if [ x`pkg-config --exists mozilla-xpcom && echo YES` = "xYES" ]; then
 		MOZILLA_INCLUDES=`pkg-config --cflags mozilla-xpcom`
 		MOZILLA_LIBS=`pkg-config --libs mozilla-xpcom`
@@ -125,9 +172,17 @@ if [ -z "${MOZILLA_INCLUDES}" -a -z "${MOZILLA_LIBS}" ]; then
 	fi
 fi
 
+if [ x`pkg-config --exists webkit-1.0 && echo YES` = "xYES" ]; then
+	echo "WebKit found, compiling webkit embedded browser support."
+	MAKE_WEBKIT=make_webkit
+else
+	echo "WebKit not found:"
+	echo "    *** WebKit embedding support will not be compiled."
+fi
+
 # Find AWT if available
 if [ -z "${AWT_LIB_PATH}" ]; then
-	if [ -d ${JAVA_HOME}/jre/lib/${AWT_ARCH} ]; then
+	if [ -f ${JAVA_HOME}/jre/lib/${AWT_ARCH}/libjawt.* ]; then
 		AWT_LIB_PATH=${JAVA_HOME}/jre/lib/${AWT_ARCH}
 		export AWT_LIB_PATH
 	else
@@ -136,7 +191,8 @@ if [ -z "${AWT_LIB_PATH}" ]; then
 	fi
 fi
 
-if [ -f ${AWT_LIB_PATH}/libjawt.so ]; then
+if [ -f ${AWT_LIB_PATH}/libjawt.* ]; then
+	echo "libjawt.so found, the SWT/AWT integration library will be compiled."
 	MAKE_AWT=make_awt
 else
 	echo "libjawt.so not found, the SWT/AWT integration library will not be compiled."
@@ -152,5 +208,5 @@ fi
 if [ "x${1}" = "xclean" ]; then
 	${MAKE_TYPE} -f $MAKEFILE clean
 else
-	${MAKE_TYPE} -f $MAKEFILE all $MAKE_GNOME $MAKE_CAIRO $MAKE_AWT $MAKE_MOZILLA ${1} ${2} ${3} ${4} ${5} ${6} ${7} ${8} ${9}
+	${MAKE_TYPE} -f $MAKEFILE all $MAKE_GNOME $MAKE_CAIRO $MAKE_AWT $MAKE_MOZILLA $MAKE_WEBKIT ${1} ${2} ${3} ${4} ${5} ${6} ${7} ${8} ${9}
 fi
