@@ -1,5 +1,5 @@
 /*******************************************************************************
- * Copyright (c) 2000, 2008 IBM Corporation and others.
+ * Copyright (c) 2000, 2010 IBM Corporation and others.
  * All rights reserved. This program and the accompanying materials
  * are made available under the terms of the Eclipse Public License v1.0
  * which accompanies this distribution, and is available at
@@ -52,6 +52,7 @@ class StyledTextRenderer {
 	StyleRange[] styles;
 	StyleRange[] stylesSet;
 	int stylesSetCount = 0;
+	boolean hasLinks, fixedPitch;
 	final static int BULLET_MARGIN = 8;
 	
 	final static boolean COMPACT_STYLES = true;
@@ -66,14 +67,20 @@ class StyledTextRenderer {
 	final static int INDENT = 1 << 2;
 	final static int JUSTIFY = 1 << 3;
 	final static int SEGMENTS = 1 << 5;
+	final static int TABSTOPS = 1 << 6;
+	final static int WRAP_INDENT = 1 << 7;
+	final static int SEGMENT_CHARS = 1 << 8;
 	
 	static class LineInfo {
 		int flags;
 		Color background;
 		int alignment;
 		int indent;
+		int wrapIndent;
 		boolean justify;
 		int[] segments;
+		char[] segmentsChars;
+		int[] tabStops;
 
 		public LineInfo() {
 		}
@@ -83,8 +90,11 @@ class StyledTextRenderer {
 				background = info.background;
 				alignment = info.alignment;
 				indent = info.indent;
+				wrapIndent = info.wrapIndent;
 				justify = info.justify;
 				segments = info.segments;
+				segmentsChars = info.segmentsChars;
+				tabStops = info.tabStops;
 			}
 		}
 	}
@@ -266,7 +276,7 @@ void clearLineStyle(int startLine, int count) {
 	for (int i = startLine; i < startLine + count; i++) {
 		LineInfo info = lines[i];
 		if (info != null) {
-			info.flags &= ~(ALIGNMENT | INDENT | JUSTIFY);
+			info.flags &= ~(ALIGNMENT | INDENT | WRAP_INDENT | JUSTIFY | TABSTOPS);
 			if (info.flags == 0) lines[i] = null;
 		}
 	}
@@ -332,7 +342,7 @@ void drawBullet(Bullet bullet, GC gc, int paintX, int paintY, int index, int lin
 	int type = bullet.type & (ST.BULLET_DOT|ST.BULLET_NUMBER|ST.BULLET_LETTER_LOWER|ST.BULLET_LETTER_UPPER);
 	switch (type) {
 		case ST.BULLET_DOT: string = "\u2022"; break;
-		case ST.BULLET_NUMBER: string = String.valueOf(index); break;
+		case ST.BULLET_NUMBER: string = String.valueOf(index + 1); break;
 		case ST.BULLET_LETTER_LOWER: string = String.valueOf((char) (index % 26 + 97)); break;
 		case ST.BULLET_LETTER_UPPER: string = String.valueOf((char) (index % 26 + 65)); break;
 	}
@@ -358,6 +368,9 @@ int drawLine(int lineIndex, int paintX, int paintY, GC gc, Color widgetBackgroun
 	Point selection = styledText.getSelection();
 	int selectionStart = selection.x - lineOffset;
 	int selectionEnd = selection.y - lineOffset;
+	if (styledText.getBlockSelection()) {
+		selectionStart = selectionEnd = 0;
+	}
 	Rectangle client = styledText.getClientArea();  
 	Color lineBackground = getLineBackground(lineIndex, null);
 	StyledTextEvent event = styledText.getLineBackgroundData(lineOffset, line);
@@ -424,7 +437,7 @@ int drawLine(int lineIndex, int paintX, int paintY, GC gc, Color widgetBackgroun
 		if (styles[i].metrics != null) {
 			if (ranges == null) ranges = layout.getRanges();
 			int start = ranges[i << 1];
-			int length = ranges[(i << 1) + 1] - start;
+			int length = ranges[(i << 1) + 1] - start + 1;
 			Point point = layout.getLocation(start, false);
 			FontMetrics metrics = layout.getLineMetrics(layout.getLineIndex(start));
 			StyleRange style = (StyleRange)((StyleRange)styles[i]).clone();
@@ -465,7 +478,7 @@ FontData[] getFontData(int style) {
 int getHeight () {
 	int defaultLineHeight = getLineHeight();
 	if (styledText.isFixedLineHeight()) {
-		return lineCount * defaultLineHeight;
+		return lineCount * defaultLineHeight + styledText.topMargin + styledText.bottomMargin;
 	}
 	int totalHeight = 0;
 	int width = styledText.getWrapWidth();
@@ -483,6 +496,45 @@ int getHeight () {
 	}
 	return totalHeight + styledText.topMargin + styledText.bottomMargin;
 }
+boolean hasLink(int offset) {
+	if (offset == -1) return false;
+	int lineIndex = content.getLineAtOffset(offset);
+	int lineOffset = content.getOffsetAtLine(lineIndex);
+	String line = content.getLine(lineIndex);
+	StyledTextEvent event = styledText.getLineStyleData(lineOffset, line);
+	if (event != null) {
+		StyleRange[] styles = event.styles;
+		if (styles != null) {
+			int[] ranges = event.ranges; 
+			if (ranges != null) {
+				for (int i = 0; i < ranges.length; i+=2) {
+					if (ranges[i] <= offset && offset < ranges[i] + ranges[i+1] && styles[i >> 1].underline && styles[i >> 1].underlineStyle == SWT.UNDERLINE_LINK) {
+						return true;
+					}
+				}
+			} else {
+				for (int i = 0; i < styles.length; i++) {
+					if (styles[i].start <= offset && offset < styles[i].start + styles[i].length && styles[i >> 1].underline && styles[i >> 1].underlineStyle == SWT.UNDERLINE_LINK) {
+						return true;
+					}
+				}
+			}
+		}
+	}  else {
+		if (ranges != null) {
+			int rangeCount = styleCount << 1;
+			int index = getRangeIndex(offset, -1, rangeCount);
+			if (index >= rangeCount) return false;
+			int rangeStart = ranges[index]; 
+			int rangeLength = ranges[index + 1];
+			StyleRange rangeStyle = styles[index >> 1];
+			if (rangeStart <= offset && offset < rangeStart + rangeLength && rangeStyle.underline && rangeStyle.underlineStyle == SWT.UNDERLINE_LINK) {
+				return true;
+			}
+		}
+	}
+	return false;
+}
 int getLineAlignment(int index, int defaultAlignment) {
 	if (lines == null) return defaultAlignment;
 	LineInfo info = lines[index];
@@ -525,6 +577,14 @@ int getLineIndent(int index, int defaultIndent) {
 	}
 	return defaultIndent;
 }
+int getLineWrapIndent(int index, int defaultWrapIndent) {
+	if (lines == null) return defaultWrapIndent;
+	LineInfo info = lines[index];
+	if (info != null && (info.flags & WRAP_INDENT) != 0) {
+		return info.wrapIndent;
+	}
+	return defaultWrapIndent;
+}
 boolean getLineJustify(int index, boolean defaultJustify) {
 	if (lines == null) return defaultJustify;
 	LineInfo info = lines[index];
@@ -533,13 +593,13 @@ boolean getLineJustify(int index, boolean defaultJustify) {
 	}
 	return defaultJustify;
 }
-int[] getLineSegments(int index, int[] defaultSegments) {
-	if (lines == null) return defaultSegments;
+int[] getLineTabStops(int index, int[] defaultTabStops) {
+	if (lines == null) return defaultTabStops;
 	LineInfo info = lines[index];
-	if (info != null && (info.flags & SEGMENTS) != 0) {
-		return info.segments;
+	if (info != null && (info.flags & TABSTOPS) != 0) {
+		return info.tabStops;
 	}
-	return defaultSegments;
+	return defaultTabStops;
 }
 int getRangeIndex(int offset, int low, int high) {
 	if (styleCount == 0) return 0;
@@ -567,6 +627,7 @@ int getRangeIndex(int offset, int low, int high) {
 	return high;
 }
 int[] getRanges(int start, int length) {
+	if (length == 0) return null;
 	int[] newRanges;
 	int end = start + length - 1;
 	if (ranges != null) {
@@ -574,7 +635,8 @@ int[] getRanges(int start, int length) {
 		int rangeStart = getRangeIndex(start, -1, rangeCount);
 		if (rangeStart >= rangeCount) return null;
 		if (ranges[rangeStart] > end) return null;
-		int rangeEnd = Math.min(rangeCount - 2, getRangeIndex(end, rangeStart - 1, rangeCount) + 1);
+		int rangeEnd = Math.min(rangeCount - 2, getRangeIndex(end, rangeStart - 1, rangeCount));
+		if (ranges[rangeEnd] > end) rangeEnd = Math.max(rangeStart, rangeEnd - 2);
 		newRanges = new int[rangeEnd - rangeStart + 2];
 		System.arraycopy(ranges, rangeStart, newRanges, 0, newRanges.length);
 	} else {
@@ -582,12 +644,13 @@ int[] getRanges(int start, int length) {
 		if (rangeStart >= styleCount) return null;
 		if (styles[rangeStart].start > end) return null;
 		int rangeEnd = Math.min(styleCount - 1, getRangeIndex(end, rangeStart - 1, styleCount));
+		if (styles[rangeEnd].start > end) rangeEnd = Math.max(rangeStart, rangeEnd - 1);
 		newRanges = new int[(rangeEnd - rangeStart + 1) << 1];
 		for (int i = rangeStart, j = 0; i <= rangeEnd; i++, j += 2) {
 			StyleRange style = styles[i];
 			newRanges[j] = style.start;
 			newRanges[j + 1] = style.length;
-		}		
+		}
 	}
 	if (start > newRanges[0]) {
 		newRanges[1] = newRanges[0] + newRanges[1] - start;
@@ -599,6 +662,7 @@ int[] getRanges(int start, int length) {
 	return newRanges;
 }
 StyleRange[] getStyleRanges(int start, int length, boolean includeRanges) {
+	if (length == 0) return null;
 	StyleRange[] newStyles;
 	int end = start + length - 1;
 	if (ranges != null) {
@@ -606,7 +670,8 @@ StyleRange[] getStyleRanges(int start, int length, boolean includeRanges) {
 		int rangeStart = getRangeIndex(start, -1, rangeCount);
 		if (rangeStart >= rangeCount) return null;
 		if (ranges[rangeStart] > end) return null;
-		int rangeEnd = Math.min(rangeCount - 2, getRangeIndex(end, rangeStart - 1, rangeCount) + 1);
+		int rangeEnd = Math.min(rangeCount - 2, getRangeIndex(end, rangeStart - 1, rangeCount));
+		if (ranges[rangeEnd] > end) rangeEnd = Math.max(rangeStart, rangeEnd - 2);
 		newStyles = new StyleRange[((rangeEnd - rangeStart) >> 1) + 1];
 		if (includeRanges) {
 			for (int i = rangeStart, j = 0; i <= rangeEnd; i += 2, j++) {
@@ -622,29 +687,27 @@ StyleRange[] getStyleRanges(int start, int length, boolean includeRanges) {
 		if (rangeStart >= styleCount) return null;
 		if (styles[rangeStart].start > end) return null;
 		int rangeEnd = Math.min(styleCount - 1, getRangeIndex(end, rangeStart - 1, styleCount));
+		if (styles[rangeEnd].start > end) rangeEnd = Math.max(rangeStart, rangeEnd - 1);
 		newStyles = new StyleRange[rangeEnd - rangeStart + 1];
 		System.arraycopy(styles, rangeStart, newStyles, 0, newStyles.length);
 	}
-	StyleRange style = newStyles[0];
-	if (start > style.start) {
-		if (!includeRanges || ranges == null) newStyles[0] = style = (StyleRange)style.clone();
-		style.length = style.start + style.length - start;
-		style.start = start;
-	}
-	style = newStyles[newStyles.length - 1];
-	if (end < style.start + style.length - 1) {
-		if (end < style.start) {
-			StyleRange[] tmp = new StyleRange[newStyles.length - 1];
-			System.arraycopy(newStyles, 0, tmp, 0, newStyles.length - 1);
-			newStyles = tmp;
-		} else {
-			if (!includeRanges || ranges == null) newStyles[newStyles.length - 1] = style = (StyleRange)style.clone();
+	if (includeRanges || ranges == null) {
+		StyleRange style = newStyles[0];
+		if (start > style.start) {
+			newStyles[0] = style = (StyleRange)style.clone();
+			style.length = style.start + style.length - start;
+			style.start = start;
+		}
+		style = newStyles[newStyles.length - 1];
+		if (end < style.start + style.length - 1) {
+			newStyles[newStyles.length - 1] = style = (StyleRange)style.clone();
 			style.length = end - style.start + 1;
 		}
 	}
 	return newStyles;
 }
 StyleRange getStyleRange(StyleRange style) {
+	if (style.underline && style.underlineStyle == SWT.UNDERLINE_LINK) hasLinks = true;
 	if (style.start == 0 && style.length == 0 && style.fontStyle == SWT.NORMAL) return style;
 	StyleRange clone = (StyleRange)style.clone();
 	clone.start = clone.length = 0;
@@ -710,28 +773,39 @@ TextLayout getTextLayout(int lineIndex, int orientation, int width, int lineSpac
 	String line = content.getLine(lineIndex);
 	int lineOffset = content.getOffsetAtLine(lineIndex);
 	int[] segments = null;
+	char[] segmentChars = null;
 	int indent = 0;
+	int wrapIndent = 0;
 	int alignment = SWT.LEFT;
 	boolean justify = false;
+	int[] tabs = {tabWidth};
 	Bullet bullet = null;
 	int[] ranges = null;
 	StyleRange[] styles = null;
 	int rangeStart = 0, styleCount = 0;
 	StyledTextEvent event = null;
 	if (styledText != null) {
+		event = styledText.getBidiSegments(lineOffset, line);
+		if (event != null) {
+			segments = event.segments;
+			segmentChars = event.segmentsChars;
+		}
 		event = styledText.getLineStyleData(lineOffset, line);
-		segments = styledText.getBidiSegments(lineOffset, line);
 		indent = styledText.indent;
+		wrapIndent = styledText.wrapIndent;
 		alignment = styledText.alignment;
 		justify = styledText.justify;
+		if (styledText.tabs != null) tabs = styledText.tabs;
 	}
 	if (event != null) {
 		indent = event.indent;
+		wrapIndent = event.wrapIndent;
 		alignment = event.alignment;
 		justify = event.justify;
 		bullet = event.bullet;
 		ranges = event.ranges;
 		styles = event.styles;
+		if (event.tabStops != null) tabs = event.tabStops;
 		if (styles != null) {
 			styleCount = styles.length;
 			if (styledText.isFixedLineHeight()) {
@@ -759,9 +833,12 @@ TextLayout getTextLayout(int lineIndex, int orientation, int width, int lineSpac
 			LineInfo info = lines[lineIndex];
 			if (info != null) {
 				if ((info.flags & INDENT) != 0) indent = info.indent;
+				if ((info.flags & WRAP_INDENT) != 0) wrapIndent = info.wrapIndent;
 				if ((info.flags & ALIGNMENT) != 0) alignment = info.alignment;
 				if ((info.flags & JUSTIFY) != 0) justify = info.justify;
 				if ((info.flags & SEGMENTS) != 0) segments = info.segments;
+				if ((info.flags & SEGMENT_CHARS) != 0) segmentChars = info.segmentsChars;
+				if ((info.flags & TABSTOPS) != 0) tabs = info.tabStops;
 			}
 		}
 		if (bulletsIndices != null) {
@@ -796,10 +873,12 @@ TextLayout getTextLayout(int lineIndex, int orientation, int width, int lineSpac
 	layout.setText(line);
 	layout.setOrientation(orientation);
 	layout.setSegments(segments);
+	layout.setSegmentsChars(segmentChars);
 	layout.setWidth(width);
 	layout.setSpacing(lineSpacing);
-	layout.setTabs(new int[]{tabWidth});
+	layout.setTabs(tabs);
 	layout.setIndent(indent);
+	layout.setWrapIndent(wrapIndent);
 	layout.setAlignment(alignment);
 	layout.setJustify(justify);
 	
@@ -960,6 +1039,7 @@ void reset() {
 	bullets = null;
 	bulletsIndices = null;
 	redrawLines = null;
+	hasLinks = false;
 }
 void reset(int startLine, int lineCount) {
 	int endLine = startLine + lineCount;
@@ -1025,6 +1105,7 @@ void setFont(Font font, int tabs) {
 	if (styledText != null) {
 		GC gc = new GC(styledText);
 		averageCharWidth = gc.getFontMetrics().getAverageCharWidth();
+		fixedPitch = gc.stringExtent("l").x == gc.stringExtent("W").x; //$NON-NLS-1$ //$NON-NLS-2$
 		gc.dispose();
 	}
 }
@@ -1087,6 +1168,16 @@ void setLineIndent(int startLine, int count, int indent) {
 		lines[i].indent = indent;
 	}
 }
+void setLineWrapIndent(int startLine, int count, int wrapIndent) {
+	if (lines == null) lines = new LineInfo[lineCount];
+	for (int i = startLine; i < startLine + count; i++) {
+		if (lines[i] == null) {
+			lines[i] = new LineInfo();
+		}
+		lines[i].flags |= WRAP_INDENT;
+		lines[i].wrapIndent = wrapIndent;
+	}
+}
 void setLineJustify(int startLine, int count, boolean justify) {
 	if (lines == null) lines = new LineInfo[lineCount];
 	for (int i = startLine; i < startLine + count; i++) {
@@ -1107,12 +1198,33 @@ void setLineSegments(int startLine, int count, int[] segments) {
 		lines[i].segments = segments;
 	}
 }
+void setLineSegmentChars(int startLine, int count, char[] segmentChars) {
+	if (lines == null) lines = new LineInfo[lineCount];
+	for (int i = startLine; i < startLine + count; i++) {
+		if (lines[i] == null) {
+			lines[i] = new LineInfo();
+		}
+		lines[i].flags |= SEGMENT_CHARS;
+		lines[i].segmentsChars = segmentChars;
+	}
+}
+void setLineTabStops(int startLine, int count, int[] tabStops) {
+	if (lines == null) lines = new LineInfo[lineCount];
+	for (int i = startLine; i < startLine + count; i++) {
+		if (lines[i] == null) {
+			lines[i] = new LineInfo();
+		}
+		lines[i].flags |= TABSTOPS;
+		lines[i].tabStops = tabStops;
+	}
+}
 void setStyleRanges (int[] newRanges, StyleRange[] newStyles) {
 	if (newStyles == null) {
 		stylesSetCount = styleCount = 0;
 		ranges = null;
 		styles = null;
 		stylesSet = null;
+		hasLinks = false;
 		return;
 	}
 	if (newRanges == null && COMPACT_STYLES) {
