@@ -1,5 +1,5 @@
 /*******************************************************************************
- * Copyright (c) 2000, 2005 IBM Corporation and others.
+ * Copyright (c) 2000, 2009 IBM Corporation and others.
  * All rights reserved. This program and the accompanying materials
  * are made available under the terms of the Eclipse Public License v1.0
  * which accompanies this distribution, and is available at
@@ -30,13 +30,18 @@ import org.eclipse.swt.accessibility.*;
  * <p>
  * IMPORTANT: This class is <em>not</em> intended to be subclassed.
  * </p>
+ *
+ * @see <a href="http://www.eclipse.org/swt/snippets/#link">Link snippets</a>
+ * @see <a href="http://www.eclipse.org/swt/examples.php">SWT Example: ControlExample</a>
+ * @see <a href="http://www.eclipse.org/swt/">Sample code and further information</a>
  * 
  * @since 3.1
+ * @noextend This class is not intended to be subclassed by clients.
  */
 public class Link extends Control {
 	String text;
 	TextLayout layout;
-	Color linkColor, linkDisabledColor;
+	Color linkColor, disabledColor;
 	Point [] offsets;
 	Point selection;
 	String [] ids;
@@ -79,11 +84,11 @@ public Link (Composite parent, int style) {
 
 /**
  * Adds the listener to the collection of listeners who will
- * be notified when the control is selected, by sending
+ * be notified when the control is selected by the user, by sending
  * it one of the messages defined in the <code>SelectionListener</code>
  * interface.
  * <p>
- * <code>widgetSelected</code> is called when the control is selected.
+ * <code>widgetSelected</code> is called when the control is selected by the user.
  * <code>widgetDefaultSelected</code> is not called.
  * </p>
  *
@@ -137,14 +142,15 @@ public Point computeSize (int wHint, int hHint, boolean changed) {
 }
 
 void createHandle(int index) {
-	state |= HANDLE;
+	state |= HANDLE | THEME_BACKGROUND;
 	handle = OS.g_object_new (display.gtk_fixed_get_type (), 0);
 	if (handle == 0) SWT.error (SWT.ERROR_NO_HANDLES);
 	OS.gtk_fixed_set_has_window (handle, true);
 	OS.GTK_WIDGET_SET_FLAGS (handle, OS.GTK_CAN_FOCUS);
 	layout = new TextLayout (display);
+	layout.setOrientation((style & SWT.RIGHT_TO_LEFT) != 0? SWT.RIGHT_TO_LEFT : SWT.LEFT_TO_RIGHT);
 	linkColor = new Color (display, LINK_FOREGROUND);
-	linkDisabledColor = new Color (display, LINK_DISABLED_FOREGROUND);
+	disabledColor = new Color (display, LINK_DISABLED_FOREGROUND);
 	offsets = new Point [0];
 	ids = new String [0];
 	mnemonics = new int [0];
@@ -162,7 +168,7 @@ void createWidget (int index) {
 void enableWidget (boolean enabled) {
 	super.enableWidget (enabled);
 	if (isDisposed ()) return;
-	TextStyle linkStyle = new TextStyle (null, enabled ? linkColor : linkDisabledColor, null);
+	TextStyle linkStyle = new TextStyle (null, enabled ? linkColor : disabledColor, null);
 	linkStyle.underline = true;
 	for (int i = 0; i < offsets.length; i++) {
 		Point point = offsets [i];
@@ -171,6 +177,10 @@ void enableWidget (boolean enabled) {
 	redraw ();
 }
 
+void fixStyle () {
+	fixStyle (handle);
+}
+
 void initAccessible () {
 	Accessible accessible = getAccessible ();
 	accessible.addAccessibleListener (new AccessibleAdapter () {
@@ -252,10 +262,13 @@ Rectangle [] getRectangles (int linkIndex) {
 	return rects;
 }
 
+int getClientWidth () {
+	return (state & ZERO_WIDTH) != 0 ? 0 : OS.GTK_WIDGET_WIDTH (handle);
+}
+
 /**
  * Returns the receiver's text, which will be an empty
- * string if it has never been set or if the receiver is
- * a <code>SEPARATOR</code> label.
+ * string if it has never been set.
  *
  * @return the receiver's text
  *
@@ -271,12 +284,14 @@ public String getText () {
 
 int /*long*/ gtk_button_press_event (int /*long*/ widget, int /*long*/ event) {
 	int /*long*/ result = super.gtk_button_press_event (widget, event);	
+	if (result != 0) return result;
 	GdkEventButton gdkEvent = new GdkEventButton ();
 	OS.memmove (gdkEvent, event, GdkEventButton.sizeof);
 	if (gdkEvent.button == 1 && gdkEvent.type == OS.GDK_BUTTON_PRESS) {
 		if (focusIndex != -1) setFocus ();
 		int x = (int) gdkEvent.x;
 		int y = (int) gdkEvent.y;
+		if ((style & SWT.MIRRORED) != 0) x = getClientWidth () - x;
 		int offset = layout.getOffset (x, y, null);
 		int oldSelectionX = selection.x;
 		int oldSelectionY = selection.y;
@@ -308,19 +323,21 @@ int /*long*/ gtk_button_press_event (int /*long*/ widget, int /*long*/ event) {
 
 int /*long*/ gtk_button_release_event (int /*long*/ widget, int /*long*/ event) {
 	int /*long*/ result = super.gtk_button_release_event (widget, event);
+	if (result != 0) return result;
 	if (focusIndex == -1) return result;
 	GdkEventButton gdkEvent = new GdkEventButton ();
 	OS.memmove (gdkEvent, event, GdkEventButton.sizeof);
 	if (gdkEvent.button == 1) {
 		int x = (int) gdkEvent.x;
 		int y = (int) gdkEvent.y;
+		if ((style & SWT.MIRRORED) != 0) x = getClientWidth () - x;
 		Rectangle [] rects = getRectangles (focusIndex);
 		for (int i = 0; i < rects.length; i++) {
 			Rectangle rect = rects [i];
 			if (rect.contains (x, y)) {
 				Event ev = new Event ();
 				ev.text = ids [focusIndex];
-				sendEvent (SWT.Selection, ev);
+				sendSelectionEvent (SWT.Selection, ev, true);
 				return result;
 			}
 		}
@@ -356,6 +373,7 @@ int /*long*/ gtk_expose_event (int /*long*/ widget, int /*long*/ eventPtr) {
 	}
 	// temporary code to disable text selection
 	selStart = selEnd = -1;
+	if ((state & DISABLED) != 0) gc.setForeground (disabledColor);
 	layout.draw (gc, 0, 0, selStart, selEnd, null, null);
 	if (hasFocus () && focusIndex != -1) {
 		Rectangle [] rects = getRectangles (focusIndex);
@@ -371,6 +389,7 @@ int /*long*/ gtk_expose_event (int /*long*/ widget, int /*long*/ eventPtr) {
 		event.y = gdkEvent.area_y;
 		event.width = gdkEvent.area_width;
 		event.height = gdkEvent.area_height;
+		if ((style & SWT.MIRRORED) != 0) event.x = getClientWidth () - event.width - event.x;
 		event.gc = gc;
 		sendEvent (SWT.Paint, event);
 		event.gc = null;
@@ -391,7 +410,7 @@ int /*long*/ gtk_key_press_event (int /*long*/ widget, int /*long*/ eventPtr) {
 		case OS.GDK_space:
 			Event event = new Event ();
 			event.text = ids [focusIndex];
-			sendEvent (SWT.Selection, event);
+			sendSelectionEvent (SWT.Selection, event, true);
 			break;
 		case OS.GDK_Tab:
 			if (focusIndex < offsets.length - 1) {
@@ -411,10 +430,12 @@ int /*long*/ gtk_key_press_event (int /*long*/ widget, int /*long*/ eventPtr) {
 
 int /*long*/ gtk_motion_notify_event (int /*long*/ widget, int /*long*/ event) {
 	int /*long*/ result = super.gtk_motion_notify_event (widget, event);
+	if (result != 0) return result;
 	GdkEventMotion gdkEvent = new GdkEventMotion ();
 	OS.memmove (gdkEvent, event, GdkEventMotion.sizeof);
 	int x = (int) gdkEvent.x;
 	int y = (int) gdkEvent.y;	
+	if ((style & SWT.MIRRORED) != 0) x = getClientWidth () - x;
 	if ((gdkEvent.state & OS.GDK_BUTTON1_MASK) != 0) {
 		int oldSelection = selection.y;
 		selection.y = layout.getOffset (x, y, null);
@@ -444,14 +465,46 @@ int /*long*/ gtk_motion_notify_event (int /*long*/ widget, int /*long*/ event) {
 	return result;
 }
 
+boolean mnemonicHit (char key) {
+	char uckey = Character.toUpperCase (key);
+	String parsedText = layout.getText();
+	for (int i = 0; i < mnemonics.length - 1; i++) {
+		if (mnemonics[i] != -1) {
+			char mnemonic = parsedText.charAt(mnemonics[i]);
+			if (uckey == Character.toUpperCase (mnemonic)) {
+				if (!setFocus ()) return false;
+				focusIndex = i;
+				redraw ();
+				return  true;
+			}
+		}
+	}
+	return false;
+}
+
+boolean mnemonicMatch (char key) {
+	char uckey = Character.toUpperCase (key);
+	String parsedText = layout.getText();
+	for (int i = 0; i < mnemonics.length - 1; i++) {
+		if (mnemonics[i] != -1) {
+			char mnemonic = parsedText.charAt(mnemonics[i]);
+			if (uckey == Character.toUpperCase (mnemonic)) { 
+				return true;
+			}
+		}
+	}
+	return false;
+}
+
+
 void releaseWidget () {
 	super.releaseWidget ();
 	if (layout != null)	layout.dispose ();
 	layout = null;
 	if (linkColor != null)	linkColor.dispose ();
 	linkColor = null;
-	if (linkDisabledColor != null) linkDisabledColor.dispose ();
-	linkDisabledColor = null;
+	if (disabledColor != null) disabledColor.dispose ();
+	disabledColor = null;
 	offsets = null;
 	ids = null;
 	mnemonics = null;
@@ -460,7 +513,7 @@ void releaseWidget () {
 
 /**
  * Removes the listener from the collection of listeners who will
- * be notified when the control is selected.
+ * be notified when the control is selected by the user.
  *
  * @param listener the listener which should no longer be notified
  *
@@ -594,7 +647,7 @@ String parse (String string) {
 	}
 	if (start < length) {
 		int tmp = parseMnemonics (buffer, start, tagStart, result);
-		int mnemonic = parseMnemonics (buffer, linkStart, index, result);
+		int mnemonic = parseMnemonics (buffer, Math.max (tagStart, linkStart), length, result);
 		if (mnemonic == -1) mnemonic = tmp;
 		mnemonics [linkIndex] = mnemonic;
 	} else {
@@ -655,9 +708,22 @@ void setFontDescription (int /*long*/ font) {
  * selected, the text field of the selection event contains either the
  * text of the hyperlink or the value of its HREF, if one was specified.
  * In the rare case of identical hyperlinks within the same string, the
- * HREF tag can be used to distinguish between them.  The string may
- * include the mnemonic character and line delimiters.
+ * HREF attribute can be used to distinguish between them.  The string may
+ * include the mnemonic character and line delimiters. The only delimiter
+ * the HREF attribute supports is the quotation mark (").
  * </p>
+ * <p>
+ * Mnemonics are indicated by an '&amp;' that causes the next
+ * character to be the mnemonic. The receiver can have a    
+ * mnemonic in the text preceding each link. When the user presses a
+ * key sequence that matches the mnemonic, focus is assigned
+ * to the link that follows the text. Mnemonics in links and in
+ * the trailing text are ignored. On most platforms,
+ * the mnemonic appears underlined but may be emphasised in a
+ * platform specific manner.  The mnemonic indicator character
+ * '&amp;' can be escaped by doubling it in the string, causing
+ * a single '&amp;' to be displayed.
+ * </p> 
  * 
  * @param string the new text
  *
@@ -678,12 +744,16 @@ public void setText (String string) {
 	focusIndex = offsets.length > 0 ? 0 : -1;
 	selection.x = selection.y = -1;
 	boolean enabled = (state & DISABLED) == 0;
-	TextStyle linkStyle = new TextStyle (null, enabled ? linkColor : linkDisabledColor, null);
+	TextStyle linkStyle = new TextStyle (null, enabled ? linkColor : disabledColor, null);
 	linkStyle.underline = true;
+	int [] bidiSegments = new int [offsets.length*2];
 	for (int i = 0; i < offsets.length; i++) {
 		Point point = offsets [i];
 		layout.setStyle (linkStyle, point.x, point.y);
+		bidiSegments[i*2] = point.x;
+		bidiSegments[i*2+1] = point.y+1;
 	}
+	layout.setSegments (bidiSegments);
 	TextStyle mnemonicStyle = new TextStyle (null, null, null);
 	mnemonicStyle.underline = true;
 	for (int i = 0; i < mnemonics.length; i++) {
@@ -695,6 +765,11 @@ public void setText (String string) {
 	redraw ();
 }
 
+void showWidget () {
+	super.showWidget ();
+	fixStyle (handle);
+}
+
 int traversalCode (int key, GdkEventKey event) {
 	if (offsets.length == 0) return 0;
 	int bits = super.traversalCode (key, event);
