@@ -1,5 +1,5 @@
 /*******************************************************************************
- * Copyright (c) 2000, 2008 IBM Corporation and others.
+ * Copyright (c) 2000, 2009 IBM Corporation and others.
  * All rights reserved. This program and the accompanying materials
  * are made available under the terms of the Eclipse Public License v1.0
  * which accompanies this distribution, and is available at
@@ -35,6 +35,7 @@ import org.eclipse.swt.events.*;
  *
  * @see <a href="http://www.eclipse.org/swt/snippets/#toolbar">ToolBar, ToolItem snippets</a>
  * @see <a href="http://www.eclipse.org/swt/">Sample code and further information</a>
+ * @noextend This class is not intended to be subclassed by clients.
  */
 public class ToolItem extends Item {
 	int /*long*/ boxHandle, arrowHandle, arrowBoxHandle, separatorHandle, labelHandle, imageHandle;
@@ -140,6 +141,12 @@ public ToolItem (ToolBar parent, int style, int index) {
  * the event object detail field contains the value <code>SWT.ARROW</code>.
  * <code>widgetDefaultSelected</code> is not called.
  * </p>
+ * <p>
+ * When the <code>SWT.RADIO</code> style bit is set, the <code>widgetSelected</code> method is
+ * also called when the receiver loses selection because another item in the same radio group 
+ * was selected by the user. During <code>widgetSelected</code> the application can use
+ * <code>getSelection()</code> to determine the current selected state of the receiver.
+ * </p>
  *
  * @param listener the listener which should be notified when the control is selected by the user,
  *
@@ -266,6 +273,19 @@ void createWidget (int index) {
 	parent.relayout ();
 }
 
+Widget [] computeTabList () {
+	if (isTabGroup ()) {
+		if (getEnabled ()) {
+			if ((style & SWT.SEPARATOR) != 0) {
+				if (control != null) return control.computeTabList();
+			} else {
+				return new Widget [] {this};
+			}
+		}
+	}
+	return new Widget [0];
+}
+
 void deregister() {
 	super.deregister ();
 	if (labelHandle != 0) display.removeWidget (labelHandle);
@@ -540,7 +560,7 @@ int /*long*/ gtk_clicked (int /*long*/ widget) {
 			selectRadio ();
 		}
 	}
-	postEvent (SWT.Selection, event);
+	sendSelectionEvent (SWT.Selection, event, false);
 	return 0;
 }
 
@@ -666,6 +686,20 @@ public boolean isEnabled () {
 	return getEnabled () && parent.isEnabled ();
 }
 
+boolean isTabGroup () {
+	ToolItem [] tabList = parent._getTabItemList ();
+	if (tabList != null) {
+		for (int i=0; i<tabList.length; i++) {
+			if (tabList [i] == this) return true;
+		}
+	}
+	if ((style & SWT.SEPARATOR) != 0) return true;
+	int index = parent.indexOf (this);
+	if (index == 0) return true;
+	ToolItem previous = parent.getItem (index - 1);
+	return (previous.getStyle () & SWT.SEPARATOR) != 0;
+}
+
 void register () {
 	super.register ();
 	if (labelHandle != 0) display.addWidget (labelHandle, this);
@@ -827,6 +861,17 @@ public void setEnabled (boolean enabled) {
 			OS.gtk_widget_hide (handle);
 			OS.gtk_widget_show (handle);
 		}
+	} else {
+		/*
+		* Bug in GTK. Starting with 2.14, if a button is disabled 
+		* through on a button press, the field which keeps track 
+		* whether the pointer is currently in the button is never updated.
+		* As a result, when it is re-enabled it automatically enters
+		* a PRELIGHT state. The fix is to set a NORMAL state.
+		*/
+		if (OS.GTK_VERSION >= OS.VERSION (2, 14, 0)) {
+			OS.gtk_widget_set_state (topHandle, OS.GTK_STATE_NORMAL);
+		}
 	}
 }
 
@@ -925,7 +970,7 @@ boolean setRadioSelection (boolean value) {
 	if ((style & SWT.RADIO) == 0) return false;
 	if (getSelection () != value) {
 		setSelection (value);
-		postEvent (SWT.Selection);
+		sendSelectionEvent (SWT.Selection);
 	}
 	return true;
 }
@@ -953,6 +998,10 @@ public void setSelection (boolean selected) {
 	OS.g_signal_handlers_unblock_matched (handle, OS.G_SIGNAL_MATCH_DATA, 0, 0, 0, 0, CLICKED);
 }
 
+boolean setTabItemFocus (boolean next) {
+	return setFocus ();
+}
+
 /**
  * Sets the receiver's text. The string may include
  * the mnemonic character.
@@ -997,8 +1046,17 @@ public void setText (String string) {
 
 /**
  * Sets the receiver's tool tip text to the argument, which
- * may be null indicating that no tool tip text should be shown.
- *
+ * may be null indicating that the default tool tip for the 
+ * control will be shown. For a control that has a default
+ * tool tip, such as the Tree control on Windows, setting
+ * the tool tip text to an empty string replaces the default,
+ * causing no tool tip text to be shown.
+ * <p>
+ * The mnemonic indicator (character '&amp;') is not displayed in a tool tip.
+ * To display a single '&amp;' in the tool tip, the character '&amp;' can be 
+ * escaped by doubling it in the string.
+ * </p>
+ * 
  * @param string the new tool tip text (or null)
  *
  * @exception SWTException <ul>
